
import torch
import torch.nn as nn
import torch.nn.functional as F
from .block import ResBlock


class SawETM(nn.Module):
    """
        Sawtooth Factorial Topic Embeddings Guided Gamma Belief Network. ICML 2021.

        Zhibin Duan, Dongsheng Wang, Bo Chen, Chaojie Wang, Wenchao Chen, Yewen Li, Jie Ren, Mingyuan Zhou.

        https://github.com/ZhibinDuan/SawETM
    """
    def __init__(self, vocab_size, num_topics_list, device='cpu', embed_size=100, hidden_size=256, pretrained_WE=None):
        super().__init__()
        # constants
        self.device = device
        self.gam_prior = torch.tensor(1.0, dtype=torch.float, device=self.device)
        self.real_min = torch.tensor(1e-30, dtype=torch.float, device=self.device)
        self.theta_max = torch.tensor(1000.0, dtype=torch.float, device=self.device)
        self.wei_shape_min = torch.tensor(1e-1, dtype=torch.float, device=self.device)
        self.wei_shape_max = torch.tensor(100.0, dtype=torch.float, device=self.device)

        # hyper-parameters
        self.num_topics_list = num_topics_list[::-1]
        self.num_hiddens_list = [hidden_size] * len(self.num_topics_list)

        assert len(self.num_topics_list) == len(self.num_hiddens_list)
        self.num_layers = len(self.num_topics_list)

        # learnable word embeddings
        if pretrained_WE is not None:
            self.rho = nn.Parameter(torch.from_numpy(pretrained_WE).float())
        else:
            self.rho = nn.Parameter(torch.empty(vocab_size, embed_size).normal_(std=0.02))

        # topic embeddings for different latent layers
        self.alpha = nn.ParameterList([])
        for n in range(self.num_layers):
            self.alpha.append(nn.Parameter(
                torch.empty(self.num_topics_list[n], embed_size).normal_(std=0.02)))

        # deterministic mapping to obtain hierarchical features
        self.h_encoder = nn.ModuleList([])
        for n in range(self.num_layers):
            if n == 0:
                self.h_encoder.append(
                    ResBlock(vocab_size, self.num_hiddens_list[n]))
            else:
                self.h_encoder.append(
                    ResBlock(self.num_hiddens_list[n - 1], self.num_hiddens_list[n]))

        # variational encoder to obtain posterior parameters
        self.q_theta = nn.ModuleList([])
        for n in range(self.num_layers):
            if n == self.num_layers - 1:
                self.q_theta.append(
                    nn.Linear(self.num_hiddens_list[n], 2 * self.num_topics_list[n]))
            else:
                self.q_theta.append(nn.Linear(
                    self.num_hiddens_list[n] + self.num_topics_list[n], 2 * self.num_topics_list[n]))

    def log_max(self, x):
        return torch.log(torch.max(x, self.real_min))

    def reparameterize(self, shape, scale, sample_num=50):
        """Returns a sample from a Weibull distribution via reparameterization.
        """
        shape = shape.unsqueeze(0).repeat(sample_num, 1, 1)
        scale = scale.unsqueeze(0).repeat(sample_num, 1, 1)
        eps = torch.rand_like(shape, dtype=torch.float, device=self.device)
        samples = scale * torch.pow(- self.log_max(1 - eps), 1 / shape)
        return torch.clamp(samples.mean(0), self.real_min.item(), self.theta_max.item())

    def kl_weibull_gamma(self, wei_shape, wei_scale, gam_shape, gam_scale):
        """Returns the Kullback-Leibler divergence between a Weibull distribution and a Gamma distribution.
        """
        euler_mascheroni_c = torch.tensor(0.5772, dtype=torch.float, device=self.device)
        t1 = torch.log(wei_shape) + torch.lgamma(gam_shape)
        t2 = - gam_shape * torch.log(wei_scale * gam_scale)
        t3 = euler_mascheroni_c * (gam_shape / wei_shape - 1) - 1
        t4 = gam_scale * wei_scale * torch.exp(torch.lgamma(1 + 1 / wei_shape))
        return (t1 + t2 + t3 + t4).sum(1).mean()

    def get_nll(self, x, x_reconstruct):
        """Returns the negative Poisson likelihood of observational count data.
        """
        log_likelihood = self.log_max(x_reconstruct) * x - torch.lgamma(1.0 + x) - x_reconstruct
        neg_log_likelihood = - torch.sum(log_likelihood, dim=1, keepdim=False).mean()
        return neg_log_likelihood

    @property
    def bottom_word_embeddings(self):
        return self.rho

    @property
    def topic_embeddings_list(self):
        return self.alpha[::-1]

    def get_phis(self):
        """Returns the factor loading matrix by utilizing sawtooth connection.
        """
        phis = []
        for n in range(self.num_layers):
            if n == 0:
                phi = torch.softmax(torch.mm(self.rho, self.alpha[n].transpose(0, 1)), dim=0)
            else:
                phi = torch.softmax(torch.mm(self.alpha[n - 1].detach(), self.alpha[n].transpose(0, 1)), dim=0)
            phis.append(phi)
        return phis

    def get_beta(self):
        beta_list = list()
        phis = self.get_phis()
        last_beta  = None

        for layer_id, phi in enumerate(phis):
            if layer_id == 0:
                last_beta = phi.T
            else:
                last_beta = torch.matmul(phi.T, last_beta)
            beta_list.append(last_beta)

        beta_list = beta_list[::-1]
        return beta_list

    def get_phi_list(self):
        phis = self.get_phis()
        phis = phis[1:]
        return [item.T for item in phis][::-1]

    def get_theta(self, x):
        hidden_feats = []
        for n in range(self.num_layers):
            if n == 0:
                hidden_feats.append(self.h_encoder[n](x))
            else:
                hidden_feats.append(self.h_encoder[n](hidden_feats[-1]))

        # =================================================================================
        # phis:
        # 5000x50
        # 50x36
        # 36x12
        # 12x2
        phis = self.get_phis()

        ks = []
        lambs = []
        thetas = []
        phi_by_theta_list = []
        for n in range(self.num_layers - 1, -1, -1):
            if n == self.num_layers - 1:
                joint_feat = hidden_feats[n]
            else:
                joint_feat = torch.cat((hidden_feats[n], phi_by_theta_list[0]), dim=1)

            k, lamb = torch.chunk(F.softplus(self.q_theta[n](joint_feat)), 2, dim=1)
            k = torch.clamp(k, self.wei_shape_min.item(), self.wei_shape_max.item())
            lamb = torch.clamp(lamb, self.real_min.item())

            if self.training:
                lamb = lamb / torch.exp(torch.lgamma(1 + 1 / k))
                theta = self.reparameterize(k, lamb, sample_num=3) if n == 0 else self.reparameterize(k, lamb)
            else:
                theta = torch.min(lamb, self.theta_max)

            phi_by_theta = torch.mm(theta, phis[n].t())
            phi_by_theta_list.insert(0, phi_by_theta)
            thetas.insert(0, theta)
            lambs.insert(0, lamb)
            ks.insert(0, k)

        if self.training:
            return ks, lambs, phi_by_theta_list, thetas
        else:
            return thetas[::-1]

    def forward(self, x):
        """Forward pass: compute the kl loss and data likelihood.
        """

        ks, lambs, phi_by_theta_list, thetas = self.get_theta(x)

        # =================================================================================
        nll = self.get_nll(x, phi_by_theta_list[0])

        kl_loss = []
        for n in range(self.num_layers):
            if n == self.num_layers - 1:
                kl_loss.append(self.kl_weibull_gamma(
                    ks[n], lambs[n], self.gam_prior, self.gam_prior))
            else:
                kl_loss.append(self.kl_weibull_gamma(
                    ks[n], lambs[n], phi_by_theta_list[n + 1], self.gam_prior))

        nelbo = nll + sum(kl_loss)

        return {'loss': nelbo}



import pytest

import sys
sys.path.append('../')

import topmost
from topmost.data import download_dataset
from topmost.data import BasicDataset
from topmost.trainers import BasicTrainer, BERTopicTrainer, FASTopicTrainer, LDAGensimTrainer, NMFGensimTrainer


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 10


def basic_model_test(model_module, dataset, num_topics, trainer=None):
    if model_module == 'BERTopic':
        trainer = BERTopicTrainer(dataset, num_topics=num_topics)

    elif model_module == 'FASTopic':
        trainer = FASTopicTrainer(dataset, num_topics=num_topics, epochs=1)
    
    elif model_module == 'LDAGensim':
        trainer = LDAGensimTrainer(dataset, num_topics=num_topics)
    
    elif model_module == 'NMFGensim':
        trainer = NMFGensimTrainer(dataset, num_topics=num_topics)

    elif model_module == 'CombinedTM':
        model = topmost.models.CombinedTM(vocab_size=dataset.vocab_size, contextual_embed_size=dataset.contextual_embed_size, num_topics=num_topics)
        trainer = BasicTrainer(model, dataset, verbose=True, epochs=1)

    else:
        model = model_module(num_topics=num_topics, vocab_size=dataset.vocab_size)
        trainer = BasicTrainer(model, dataset, verbose=True, epochs=1)

    trainer.train()

    assert trainer.get_beta().shape == (num_topics, dataset.vocab_size)

    train_theta, test_theta = trainer.export_theta()

    if model_module != 'BERTopic':
        assert train_theta.shape == (len(dataset.train_texts), num_topics)
        assert test_theta.shape == (len(dataset.test_texts), num_topics)


def test_models(cache_path, num_topics):
    download_dataset("20NG", cache_path=cache_path)
    dataset = BasicDataset(f"{cache_path}/20NG")

    model_info = [
        topmost.ProdLDA,
        topmost.ETM,
        topmost.DecTM,
        topmost.NSTM,
        topmost.TSCTM,
        topmost.ECRTM,
        'BERTopic',
        'FASTopic',
        'LDAGensim',
        'NMFGensim'
    ]

    for model_module in model_info:
        print(model_module)
        basic_model_test(model_module, dataset, num_topics)

    dataset = BasicDataset(f"{cache_path}/20NG", contextual_embed=True)
    basic_model_test("CombinedTM", dataset, num_topics)



import numpy as np
from sklearn.feature_extraction.text import CountVectorizer
from collections import Counter
from tqdm import tqdm


def _diversity(top_words):
    K = len(top_words)
    T = len(top_words[0].split())
    vectorizer = CountVectorizer(tokenizer=lambda x: x.split())
    counter = vectorizer.fit_transform(top_words).toarray()

    TF = counter.sum(axis=0)
    TD = (TF == 1).sum() / (K * T)

    return TD


def multiaspect_diversity(top_words, _type="TD"):
    TD_list = list()
    for level_top_words in top_words:
        TD = _diversity(level_top_words, _type)
        TD_list.append(TD)

    return np.mean(TD_list)


def _time_slice_diversity(topics, time_vocab):
    num_associated_words = 0.
    T = len(topics[0].split())
    flatten_topic_words = [word for topic_words in topics for word in topic_words.split()]
    counter = Counter(flatten_topic_words)

    # for word in np.sort(list(set(flatten_topic_words))):
    for word in np.sort(flatten_topic_words):
        if (counter[word] == 1) and word in time_vocab:
            num_associated_words += 1

    return num_associated_words / (len(topics) * T)


def dynamic_diversity(top_words, train_bow, train_times, vocab, verbose=False):
    TD_list = list()

    time_idx = np.sort(np.unique(train_times))

    for time in tqdm(time_idx):
        doc_idx = np.where(train_times == time)[0]
        time_vocab_idx = np.nonzero(train_bow[doc_idx].sum(0))[0]
        time_vocab = np.asarray(vocab)[time_vocab_idx]

        topics = top_words[time]
        TD_list.append(_time_slice_diversity(topics, time_vocab))

    if verbose:
        print(f"dynamic TD list: {TD_list}")

    return np.mean(TD_list)



import numpy as np
from sklearn.svm import SVC
from sklearn.metrics import f1_score, accuracy_score
from collections import defaultdict


def _cls(train_theta, test_theta, train_labels, test_labels, classifier='SVM', gamma='scale'):
    if classifier == 'SVM':
        clf = SVC(gamma=gamma)
    else:
        raise NotImplementedError

    clf.fit(train_theta, train_labels)
    preds = clf.predict(test_theta)
    results = {
        'acc': accuracy_score(test_labels, preds),
        'macro-F1': f1_score(test_labels, preds, average='macro')
    }
    return results


def crosslingual_cls(
    train_theta_en,
    train_theta_cn,
    test_theta_en,
    test_theta_cn,
    train_labels_en,
    train_labels_cn,
    test_labels_en,
    test_labels_cn,
    classifier="SVM",
    gamma="scale"
):
    intra_en = _cls(train_theta_en, test_theta_en, train_labels_en, test_labels_en, classifier, gamma)
    intra_cn = _cls(train_theta_cn, test_theta_cn, train_labels_cn, test_labels_cn, classifier, gamma)

    cross_en = _cls(train_theta_cn, test_theta_en, train_labels_cn, test_labels_en, classifier, gamma)
    cross_cn = _cls(train_theta_en, test_theta_cn, train_labels_en, test_labels_cn, classifier, gamma)

    return {
        'intra_en': intra_en,
        'intra_cn': intra_cn,
        'cross_en': cross_en,
        'cross_cn': cross_cn
    }


def hierarchical_cls(train_theta, test_theta, train_labels, test_labels, classifier='SVM', gamma='scale'):
    num_layer = len(train_theta)
    results = defaultdict(list)

    for layer in range(num_layer):
        layer_results = _cls(train_theta[layer], test_theta[layer], train_labels, test_labels, classifier, gamma)

        for key in layer_results:
            results[key].append(layer_results[key])

    for key in results:
        results[key] = np.mean(results[key])

    return results



from bertopic import BERTopic


class BERTopicTrainer:
    def __init__(self, dataset, num_topics=50, num_top_words=15):
        self.model = BERTopic(nr_topics=num_topics, top_n_words=num_top_words)
        self.dataset = dataset

    def train(self):
        self.model.fit_transform(self.dataset.train_texts)
        top_words = self.get_top_words()
        train_theta = self.test(self.dataset.train_texts)

        return top_words, train_theta

    def test(self, texts):
        theta, _ = self.model.approximate_distribution(texts)
        return theta

    def get_beta(self):
        # NOTE: beta is modeled as unnormalized c-tf_idf.
        beta = self.model.c_tf_idf_.toarray()
        return beta

    def get_top_words(self):
        top_words = list()
        for item in self.model.get_topics().values():
            top_words.append(' '.join([x[0] for x in item]))
        return top_words

    def export_theta(self):
        train_theta = self.test(self.dataset.train_texts)
        test_theta = self.test(self.dataset.test_texts)
        return train_theta, test_theta



import torch
import torch.nn as nn
import torch.nn.functional as F


class ETC(nn.Module):
    def __init__(self, num_times, temperature, weight_neg, weight_pos):
        super().__init__()
        self.num_times = num_times
        self.weight_neg = weight_neg
        self.weight_pos = weight_pos
        self.temperature = temperature

    def forward(self, topic_embeddings):
        loss = 0.
        loss_neg = 0.
        loss_pos = 0.

        for t in range(self.num_times):
            loss_neg += self.compute_loss(topic_embeddings[t], topic_embeddings[t], self.temperature, self_contrast=True)

        for t in range(1, self.num_times):
            loss_pos += self.compute_loss(topic_embeddings[t], topic_embeddings[t - 1].detach(), self.temperature, self_contrast=False, only_pos=True)

        loss_neg *= (self.weight_neg / self.num_times)
        loss_pos *= (self.weight_pos / (self.num_times - 1))
        loss = loss_neg + loss_pos

        return loss

    def compute_loss(self, anchor_feature, contrast_feature, temperature, self_contrast=False, only_pos=False, all_neg=False):
        # KxK
        anchor_dot_contrast = torch.div(
            torch.matmul(F.normalize(anchor_feature, dim=1), F.normalize(contrast_feature, dim=1).T),
            temperature
        )

        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        pos_mask = torch.eye(anchor_dot_contrast.shape[0]).to(anchor_dot_contrast.device)

        if self_contrast is False:
            if only_pos is False:
                if all_neg is True:
                    exp_logits = torch.exp(logits)
                    sum_exp_logits = exp_logits.sum(1)
                    log_prob = -torch.log(sum_exp_logits + 1e-12)

                    mean_log_prob = -log_prob.sum() / (logits.shape[0] * logits.shape[1])
            else:
                # only pos
                mean_log_prob = -(logits * pos_mask).sum() / pos_mask.sum()
        else:
            # self contrast: push away from each other in the same time slice.
            exp_logits = torch.exp(logits) * (1 - pos_mask)
            sum_exp_logits = exp_logits.sum(1)
            log_prob = -torch.log(sum_exp_logits + 1e-12)

            mean_log_prob = -log_prob.sum() / (1 - pos_mask).sum()

        return mean_log_prob



import torch
import torch.nn as nn
import torch.nn.functional as F


class TAMI(nn.Module):
    '''
        InfoCTM: A Mutual Information Maximization Perspective of Cross-lingual Topic Modeling. AAAI 2023

        Xiaobao Wu, Xinshuai Dong, Thong Nguyen, Chaoqun Liu, Liangming Pan, Anh Tuan Luu
    '''
    def __init__(self, temperature, weight_MI, pos_threshold, trans_e2c, pretrain_word_embeddings_en, pretrain_word_embeddings_cn):
        super().__init__()
        self.temperature = temperature
        self.weight_MI = weight_MI
        self.pos_threshold = pos_threshold
        self.pretrain_word_embeddings_en = pretrain_word_embeddings_en
        self.pretrain_word_embeddings_cn = pretrain_word_embeddings_cn

        self.trans_e2c = torch.as_tensor(trans_e2c).float()
        self.trans_e2c = nn.Parameter(self.trans_e2c, requires_grad=False)
        self.trans_c2e = self.trans_e2c.T

        pos_trans_mask_en, pos_trans_mask_cn, neg_trans_mask_en, neg_trans_mask_cn = self.compute_pos_neg(pretrain_word_embeddings_en, pretrain_word_embeddings_cn, self.trans_e2c, self.trans_c2e)
        self.pos_trans_mask_en = nn.Parameter(pos_trans_mask_en, requires_grad=False)
        self.pos_trans_mask_cn = nn.Parameter(pos_trans_mask_cn, requires_grad=False)
        self.neg_trans_mask_en = nn.Parameter(neg_trans_mask_en, requires_grad=False)
        self.neg_trans_mask_cn = nn.Parameter(neg_trans_mask_cn, requires_grad=False)

    def build_CVL_mask(self, embeddings):
        norm_embed = F.normalize(embeddings)
        cos_sim = torch.matmul(norm_embed, norm_embed.T)
        pos_mask = (cos_sim >= self.pos_threshold).float()
        return pos_mask

    def translation_mask(self, mask, trans_dict_matrix):
        # V1 x V2
        trans_mask = torch.matmul(mask, trans_dict_matrix)
        return trans_mask

    def compute_pos_neg(self, pretrain_word_embeddings_en, pretrain_word_embeddings_cn, trans_e2c, trans_c2e):
        # Ve x Ve
        pos_mono_mask_en = self.build_CVL_mask(torch.as_tensor(pretrain_word_embeddings_en))
        # Vc x Vc
        pos_mono_mask_cn = self.build_CVL_mask(torch.as_tensor(pretrain_word_embeddings_cn))

        # Ve x Vc
        pos_trans_mask_en = self.translation_mask(pos_mono_mask_en, trans_e2c)
        pos_trans_mask_cn = self.translation_mask(pos_mono_mask_cn, trans_c2e)

        neg_trans_mask_en = (pos_trans_mask_en <= 0).float()
        neg_trans_mask_cn = (pos_trans_mask_cn <= 0).float()

        return pos_trans_mask_en, pos_trans_mask_cn, neg_trans_mask_en, neg_trans_mask_cn

    def MutualInfo(self, anchor_feature, contrast_feature, mask, neg_mask):
        anchor_dot_contrast = torch.div(
            torch.matmul(F.normalize(anchor_feature, dim=1), F.normalize(contrast_feature, dim=1).T),
            self.temperature
        )
        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        exp_logits = torch.exp(logits) * neg_mask
        sum_exp_logits = exp_logits.sum(1, keepdim=True)

        log_prob = logits - torch.log(sum_exp_logits + torch.exp(logits) + 1e-10)
        mean_log_prob = -(mask * log_prob).sum()
        return mean_log_prob

    def forward(self, fea_en, fea_cn):
        loss_TAMI = self.MutualInfo(fea_en, fea_cn, self.pos_trans_mask_en, self.neg_trans_mask_en)
        loss_TAMI += self.MutualInfo(fea_cn, fea_en, self.pos_trans_mask_cn, self.neg_trans_mask_cn)

        loss_TAMI = loss_TAMI / (self.pos_trans_mask_en.sum() + self.pos_trans_mask_cn.sum())

        loss_TAMI = self.weight_MI * loss_TAMI
        return loss_TAMI



from .topic_diversity import _diversity
from .topic_diversity import multiaspect_diversity
from .topic_diversity import dynamic_diversity

from .clustering import _clustering
from .clustering import hierarchical_clustering

from .classification import _cls
from .classification import crosslingual_cls
from .classification import hierarchical_cls

from .topic_coherence import _coherence
from .topic_coherence import dynamic_coherence
from .hierarchy_quality import hierarchy_quality



import torch
from torch.utils.data import DataLoader
import numpy as np
import scipy.io
import scipy.sparse
from scipy.sparse import issparse
from sentence_transformers import SentenceTransformer
from topmost.preprocess import Preprocess
from . import file_utils
from typing import List, Tuple, Union, Mapping, Any, Callable, Iterable


class DocEmbedModel:
    def __init__(
            self,
            model: Union[str, callable]="all-MiniLM-L6-v2",
            device: str='cpu',
            verbose: bool=False
        ):
        self.verbose = verbose

        if isinstance(model, str): 
            self.model = SentenceTransformer(model, device=device)
        else:
            self.model = model

    def encode(self,
               docs:List[str],
               convert_to_tensor: bool=False
            ):

        embeddings = self.model.encode(
                        docs,
                        convert_to_tensor=convert_to_tensor,
                        show_progress_bar=self.verbose
                    )
        return embeddings


class RawDataset:
    def __init__(self,
                 docs,
                 preprocess=None,
                 batch_size=200,
                 device='cpu',
                 as_tensor=True,
                 contextual_embed=False,
                 pretrained_WE=False,
                 doc_embed_model="all-MiniLM-L6-v2",
                 embed_model_device=None,
                 verbose=False
                ):

        if preprocess is None:
            preprocess = Preprocess(verbose=verbose)

        rst = preprocess.preprocess(docs, pretrained_WE=pretrained_WE)
        self.train_data = rst["train_bow"]
        self.train_texts = rst["train_texts"]
        self.vocab = rst["vocab"]
        if issparse(self.train_data):
            self.train_data = self.train_data.toarray()

        self.vocab_size = len(self.vocab)

        if contextual_embed:
            if embed_model_device is None:
                embed_model_device = device

            if isinstance(doc_embed_model, str):
                self.doc_embedder = DocEmbedModel(doc_embed_model, embed_model_device, verbose=verbose)
            else:
                self.doc_embedder = doc_embed_model

            self.train_contextual_embed = self.doc_embedder.encode(docs)
            self.contextual_embed_size = self.train_contextual_embed.shape[1]

        if as_tensor:
            if contextual_embed:
                self.train_data = np.concatenate((self.train_data, self.train_contextual_embed), axis=1)

            self.train_data = torch.from_numpy(self.train_data).float().to(device)
            self.train_dataloader = DataLoader(self.train_data, batch_size=batch_size, shuffle=True)


class BasicDataset:
    def __init__(self,
                 dataset_dir,
                 batch_size=200,
                 read_labels=False,
                 as_tensor=True,
                 contextual_embed=False,
                 doc_embed_model="all-MiniLM-L6-v2",
                 device='cpu'
                ):
        # train_bow: NxV
        # test_bow: Nxv
        # word_emeddings: VxD
        # vocab: V, ordered by word id.

        self.load_data(dataset_dir, read_labels)
        self.vocab_size = len(self.vocab)

        print("train_size: ", self.train_bow.shape[0])
        print("test_size: ", self.test_bow.shape[0])
        print("vocab_size: ", self.vocab_size)
        print("average length: {:.3f}".format(self.train_bow.sum(1).sum() / self.train_bow.shape[0]))

        if contextual_embed:
            self.doc_embedder = DocEmbedModel(doc_embed_model, device)
            self.train_contextual_embed = self.doc_embedder.encode(self.train_texts)
            self.test_contextual_embed = self.doc_embedder.encode(self.test_texts)

            self.contextual_embed_size = self.train_contextual_embed.shape[1]

        if as_tensor:
            if not contextual_embed:
                self.train_data = self.train_bow
                self.test_data = self.test_bow
            else:
                self.train_data = np.concatenate((self.train_bow, self.train_contextual_embed), axis=1)
                self.test_data = np.concatenate((self.test_bow, self.test_contextual_embed), axis=1)

            self.train_data = torch.from_numpy(self.train_data).to(device)
            self.test_data = torch.from_numpy(self.test_data).to(device)

            self.train_dataloader = DataLoader(self.train_data, batch_size=batch_size, shuffle=True)
            self.test_dataloader = DataLoader(self.test_data, batch_size=batch_size, shuffle=False)

    def load_data(self, path, read_labels):

        self.train_bow = scipy.sparse.load_npz(f'{path}/train_bow.npz').toarray().astype('float32')
        self.test_bow = scipy.sparse.load_npz(f'{path}/test_bow.npz').toarray().astype('float32')
        self.pretrained_WE = scipy.sparse.load_npz(f'{path}/word_embeddings.npz').toarray().astype('float32')

        self.train_texts = file_utils.read_text(f'{path}/train_texts.txt')
        self.test_texts = file_utils.read_text(f'{path}/test_texts.txt')

        if read_labels:
            self.train_labels = np.loadtxt(f'{path}/train_labels.txt', dtype=int)
            self.test_labels = np.loadtxt(f'{path}/test_labels.txt', dtype=int)

        self.vocab = file_utils.read_text(f'{path}/vocab.txt')



import torch
from torch import nn
import torch.nn.functional as F
from . import utils


class CDDecoder(nn.Module):
    def __init__(self, num_layers, vocab_size, bias_p, bias_topk):
        super().__init__()
        self.num_layers = num_layers
        self.bias_p = bias_p
        self.bias_topk = bias_topk

        self.decoder_bn_list = nn.ModuleList([nn.BatchNorm1d(vocab_size, affine=False) for _ in range(num_layers)])

        self.bias_vectors = nn.ParameterList([])
        for _ in range(num_layers):
            bias_vector = nn.Parameter(nn.init.xavier_uniform_(torch.empty(1, vocab_size)))
            self.bias_vectors.append(bias_vector)

    def forward(self, input_bow, theta_list, beta_list):
        topk_bias_list = list()
        all_recon_loss = 0.

        for layer_id in range(self.num_layers):
            topk_bias = utils.get_topk_tensor(beta_list[layer_id], topk=self.bias_topk).sum(0)
            topk_bias = topk_bias.detach()
            topk_bias_list.append(topk_bias)

        for layer_id in range(self.num_layers):
            topk_bias = 0.
            # previous layer
            if layer_id > 0:
                topk_bias += topk_bias_list[layer_id - 1]

            # next layer
            if layer_id < self.num_layers - 1:
                topk_bias += topk_bias_list[layer_id + 1]

            topk_mask = (topk_bias > 0).float()
            bias = self.bias_p * topk_bias * topk_mask + self.bias_vectors[layer_id] * (1 - topk_mask)

            recon = self.decoder_bn_list[layer_id](torch.matmul(theta_list[layer_id], beta_list[layer_id]))
            recon = recon + bias
            recon = F.softmax(recon, dim=-1)
            recon_loss = -(input_bow * (recon + 1e-12).log()).sum(axis=1)
            recon_loss = recon_loss.mean()
            all_recon_loss += recon_loss

        all_recon_loss /= self.num_layers

        return all_recon_loss



from . import models
from . import data
from . import eva
from . import trainers
from . import preprocess

from .preprocess.preprocess import Preprocess

# data
from .data.basic_dataset import BasicDataset
from .data.basic_dataset import RawDataset
from .data.crosslingual_dataset import CrosslingualDataset
from .data.dynamic_dataset import DynamicDataset
from .data.download import download_dataset


# trainers
from .trainers.basic.basic_trainer import BasicTrainer
from .trainers.basic.BERTopic_trainer import BERTopicTrainer
from .trainers.basic.FASTopic_trainer import FASTopicTrainer
from .trainers.basic.LDA_trainer import LDAGensimTrainer
from .trainers.basic.LDA_trainer import LDASklearnTrainer
from .trainers.basic.NMF_trainer import NMFGensimTrainer
from .trainers.basic.NMF_trainer import NMFSklearnTrainer

from .trainers.crosslingual.crosslingual_trainer import CrosslingualTrainer
from .trainers.dynamic.dynamic_trainer import DynamicTrainer

from .trainers.dynamic.DTM_trainer import DTMTrainer

from .trainers.hierarchical.hierarchical_trainer import HierarchicalTrainer
from .trainers.hierarchical.HDP_trainer import HDPGensimTrainer

# models
from .models.basic.ProdLDA import ProdLDA
from .models.basic.CombinedTM import CombinedTM
from .models.basic.DecTM import DecTM
from .models.basic.ETM import ETM
from .models.basic.NSTM.NSTM import NSTM
from .models.basic.TSCTM.TSCTM import TSCTM
from .models.basic.ECRTM.ECRTM import ECRTM

from .models.crosslingual.NMTM import NMTM
from .models.crosslingual.InfoCTM.InfoCTM import InfoCTM

from .models.dynamic.DETM import DETM
from .models.dynamic.CFDTM.CFDTM import CFDTM

from .models.hierarchical.SawETM.SawETM import SawETM
from .models.hierarchical.HyperMiner.HyperMiner import HyperMiner
from .models.hierarchical.TraCo.TraCo import TraCo



from setuptools import setup, find_packages


with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

with open('README.rst') as f:
    readme = f.read()

VERSION = '1.0.0'
DESCRIPTION = 'Topmost: A Topic Modeling System Tookit'
LONG_DESCRIPTION = """
    [TopMost](https://github.com/bobxwu/topmost) provides complete lifecycles of topic modeling, including datasets, models, training, and evaluations.
    It covers the most popular topic modeling scenarios, like basic, dynamic, hierarchical, and cross-lingual topic modeling.
"""


# Setting up
setup(
        name="topmost",
        version=VERSION,
        author="Xiaobao Wu",
        author_email="xiaobao002@e.ntu.edu.sg",
        description=DESCRIPTION,
        long_description=readme,
        long_description_content_type="text/x-rst",
        # long_description_content_type="text/markdown",
        url='https://github.com/bobxwu/topmost',
        packages=find_packages(),
        # packages=find_packages(include=['topmost', 'topmost.*']),
        license="Apache 2.0 License",
        install_requires=requirements,
        keywords=['toolkit', 'topic model', 'neural topic model'],
        include_package_data=True,
        test_suite='tests',
        classifiers= [
            'Development Status :: 3 - Alpha',
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)



from fastopic import FASTopic
from topmost.utils.logger import Logger


logger = Logger("WARNING")


class FASTopicTrainer:
    def __init__(self,
                 dataset,
                 num_topics=50,
                 num_top_words=15,
                 preprocess=None,
                 epochs=200,
                 DT_alpha=3.0,
                 TW_alpha=2.0,
                 theta_temp=1.0,
                 verbose=False
                ):
        self.dataset = dataset
        self.num_top_words = num_top_words

        self.model = FASTopic(num_topics=num_topics,
                              preprocess=preprocess,
                              num_top_words=num_top_words,
                              DT_alpha=DT_alpha,
                              TW_alpha=TW_alpha,
                              theta_temp=theta_temp,
                              verbose=verbose
                            )

        self.epochs = epochs

        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

    def train(self):
        return self.model.fit_transform(self.dataset.train_texts, epochs=self.epochs)

    def test(self, texts):
        theta = self.model.transform(texts)
        return theta

    def get_beta(self):
        beta = self.model.get_beta()
        return beta

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        return self.model.get_top_words(num_top_words)

    def export_theta(self):
        train_theta = self.test(self.dataset.train_texts)
        test_theta = self.test(self.dataset.test_texts)
        return train_theta, test_theta



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .ETC import ETC
from .UWE import UWE
from topmost.models.Encoder import MLPEncoder


class CFDTM(nn.Module):
    '''
    Modeling Dynamic Topics in Chain-Free Fashion by Evolution-Tracking Contrastive Learning and Unassociated Word Exclusion. ACL 2024 Findings

    Xiaobao Wu, Xinshuai Dong, Liangming Pan, Thong Nguyen, Anh Tuan Luu.
    '''

    def __init__(self,
                 vocab_size,
                 train_time_wordfreq,
                 num_times,
                 pretrained_WE=None,
                 num_topics=50,
                 en_units=100,
                 temperature=0.1,
                 beta_temp=1.0,
                 weight_neg=1.0e+7,
                 weight_pos=1.0e+1,
                 weight_UWE=1.0e+3,
                 neg_topk=15,
                 dropout=0.,
                 embed_size=200
                ):
        super().__init__()

        self.num_topics = num_topics
        self.beta_temp = beta_temp
        self.train_time_wordfreq = train_time_wordfreq
        self.encoder = MLPEncoder(vocab_size, num_topics, en_units, dropout)

        self.a = 1 * np.ones((1, num_topics)).astype(np.float32)
        self.mu2 = nn.Parameter(torch.as_tensor((np.log(self.a).T - np.mean(np.log(self.a), 1)).T))
        self.var2 = nn.Parameter(torch.as_tensor((((1.0 / self.a) * (1 - (2.0 / num_topics))).T + (1.0 / (num_topics * num_topics)) * np.sum(1.0 / self.a, 1)).T))

        self.mu2.requires_grad = False
        self.var2.requires_grad = False

        self.decoder_bn = nn.BatchNorm1d(vocab_size, affine=False)

        if pretrained_WE is None:
            self.word_embeddings  = nn.init.trunc_normal_(torch.empty(vocab_size, embed_size), std=0.1)
            self.word_embeddings = nn.Parameter(F.normalize(self.word_embeddings))

        else:
            self.word_embeddings = nn.Parameter(torch.from_numpy(pretrained_WE).float())

        # topic_embeddings: TxKxD
        self.topic_embeddings = nn.init.xavier_normal_(torch.zeros(num_topics, self.word_embeddings.shape[1])).repeat(num_times, 1, 1)
        self.topic_embeddings = nn.Parameter(self.topic_embeddings)

        self.ETC = ETC(num_times, temperature, weight_neg, weight_pos)
        self.UWE = UWE(self.ETC, num_times, temperature, weight_UWE, neg_topk)

    def get_beta(self):
        dist = self.pairwise_euclidean_dist(F.normalize(self.topic_embeddings, dim=-1), F.normalize(self.word_embeddings, dim=-1))
        beta = F.softmax(-dist / self.beta_temp, dim=1)

        return beta

    def pairwise_euclidean_dist(self, x, y):
        cost = torch.sum(x ** 2, axis=-1, keepdim=True) + torch.sum(y ** 2, axis=-1) - 2 * torch.matmul(x, y.t())
        return cost

    def get_theta(self, x, times=None):
        theta, mu, logvar = self.encoder(x)
        if self.training:
            return theta, mu, logvar

        return theta

    def get_KL(self, mu, logvar):
        var = logvar.exp()
        var_division = var / self.var2
        diff = mu - self.mu2
        diff_term = diff * diff / self.var2
        logvar_division = self.var2.log() - logvar
        KLD = 0.5 * ((var_division + diff_term + logvar_division).sum(axis=1) - self.num_topics)

        return KLD.mean()

    def get_NLL(self, theta, beta, x, recon_x=None):
        if recon_x is None:
            recon_x = self.decode(theta, beta)
        recon_loss = -(x * recon_x.log()).sum(axis=1)

        return recon_loss

    def decode(self, theta, beta):
        d1 = F.softmax(self.decoder_bn(torch.bmm(theta.unsqueeze(1), beta).squeeze(1)), dim=-1)
        return d1

    def forward(self, x, times):
        loss = 0.

        theta, mu, logvar = self.get_theta(x)
        kl_theta = self.get_KL(mu, logvar)

        loss += kl_theta

        beta = self.get_beta()
        time_index_beta = beta[times]
        recon_x = self.decode(theta, time_index_beta)
        NLL = self.get_NLL(theta, time_index_beta, x, recon_x)
        NLL = NLL.mean()
        loss += NLL

        loss_ETC = self.ETC(self.topic_embeddings)
        loss += loss_ETC

        loss_UWE = self.UWE(self.train_time_wordfreq, beta, self.topic_embeddings, self.word_embeddings)
        loss += loss_UWE

        rst_dict = {
            'loss': loss,
        }

        return rst_dict



import pytest

import sys
sys.path.append('../')

import topmost
from topmost import download_dataset
from topmost import DynamicDataset, DynamicTrainer


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 5


def dynamic_model_test(model_module, dataset, num_topics):

    if model_module == topmost.models.DETM:
        model = model_module(num_times=dataset.num_times, train_size=dataset.train_size, num_topics=num_topics, vocab_size=dataset.vocab_size, train_time_wordfreq=dataset.train_time_wordfreq)
    elif model_module == topmost.models.CFDTM:
        model = model_module(num_times=dataset.num_times, pretrained_WE=dataset.pretrained_WE, num_topics=num_topics, vocab_size=dataset.vocab_size, train_time_wordfreq=dataset.train_time_wordfreq)

    trainer = DynamicTrainer(model, dataset, verbose=True, epochs=1)
    trainer.train()

    beta = trainer.get_beta()
    assert beta.shape == (dataset.num_times, num_topics, dataset.vocab_size)

    train_theta, test_theta = trainer.export_theta()
    assert train_theta.shape == (len(dataset.train_texts), num_topics)
    assert test_theta.shape == (len(dataset.test_texts), num_topics)


def test_models(cache_path, num_topics):
    download_dataset("NYT", cache_path=cache_path)
    dataset = DynamicDataset(f"{cache_path}/NYT", as_tensor=True)

    model_info = [
        topmost.DETM,
        topmost.CFDTM,
    ]

    for model_module in model_info:
        print(model_module)
        dynamic_model_test(model_module, dataset, num_topics)





import torch
from torch import nn
import torch.nn.functional as F
from .auto_diff_sinkhorn import sinkhorn_loss


class NSTM(nn.Module):
    '''
        Neural Topic Model via Optimal Transport. ICLR 2021

        He Zhao, Dinh Phung, Viet Huynh, Trung Le, Wray Buntine.
    '''
    def __init__(self, vocab_size, num_topics=50, en_units=200, dropout=0.25, pretrained_WE=None, train_WE=True, embed_size=200, recon_loss_weight=0.07, sinkhorn_alpha=20):
        super().__init__()

        self.recon_loss_weight = recon_loss_weight
        self.sinkhorn_alpha = sinkhorn_alpha

        self.e1 = nn.Linear(vocab_size, en_units)
        self.e2 = nn.Linear(en_units, num_topics)
        self.e_dropout = nn.Dropout(dropout)
        self.mean_bn = nn.BatchNorm1d(num_topics)
        self.mean_bn.weight.requires_grad = False

        if pretrained_WE is not None:
            self.word_embeddings = nn.Parameter(torch.from_numpy(pretrained_WE).float())
        else:
            self.word_embeddings = nn.Parameter(torch.randn((vocab_size, embed_size)))

        self.word_embeddings.requires_grad = train_WE

        self.topic_embeddings = torch.empty((num_topics, self.word_embeddings.shape[1]))
        nn.init.trunc_normal_(self.topic_embeddings, std=0.1)
        self.topic_embeddings = nn.Parameter(self.topic_embeddings)

    def get_beta(self):
        word_embedding_norm = F.normalize(self.word_embeddings)
        topic_embedding_norm = F.normalize(self.topic_embeddings)
        beta = torch.matmul(topic_embedding_norm, word_embedding_norm.T)
        return beta

    def get_theta(self, input):
        theta = F.relu(self.e1(input))
        theta = self.e_dropout(theta)
        theta = self.mean_bn(self.e2(theta))
        theta = F.softmax(theta, dim=-1)
        return theta

    def forward(self, input):
        theta = self.get_theta(input)
        beta = self.get_beta()
        M = 1 - beta
        sh_loss = sinkhorn_loss(M, theta.T, F.softmax(input, dim=-1).T, lambda_sh=self.sinkhorn_alpha)
        recon = F.softmax(torch.matmul(theta, beta), dim=-1)
        recon_loss = -(input * recon.log()).sum(axis=1)

        loss = self.recon_loss_weight * recon_loss + sh_loss
        loss = loss.mean()
        return {'loss': loss}



import torch
import torch.nn as nn
import torch.nn.functional as F


class MLPEncoder(nn.Module):
    def __init__(self, vocab_size, num_topic, hidden_dim, dropout):
        super().__init__()

        self.fc11 = nn.Linear(vocab_size, hidden_dim)
        self.fc12 = nn.Linear(hidden_dim, hidden_dim)
        self.fc21 = nn.Linear(hidden_dim, num_topic)
        self.fc22 = nn.Linear(hidden_dim, num_topic)

        self.fc1_drop = nn.Dropout(dropout)
        self.z_drop = nn.Dropout(dropout)

        self.mean_bn = nn.BatchNorm1d(num_topic, affine=True)
        self.mean_bn.weight.requires_grad = False
        self.logvar_bn = nn.BatchNorm1d(num_topic, affine=True)
        self.logvar_bn.weight.requires_grad = False

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return mu + (eps * std)
        else:
            return mu

    def forward(self, x):
        e1 = F.softplus(self.fc11(x))
        e1 = F.softplus(self.fc12(e1))
        e1 = self.fc1_drop(e1)
        mu = self.mean_bn(self.fc21(e1))
        logvar = self.logvar_bn(self.fc22(e1))
        theta = self.reparameterize(mu, logvar)
        theta = F.softmax(theta, dim=1)
        theta = self.z_drop(theta)
        return theta, mu, logvar



import gensim
import numpy as np
from gensim.models import ldaseqmodel
from tqdm import tqdm
import datetime
from multiprocessing.pool import Pool
from topmost.utils import _utils
from topmost.utils.logger import Logger


logger = Logger("WARNING")


def work(arguments):
    model, docs = arguments
    theta_list = list()
    for doc in tqdm(docs):
        theta_list.append(model[doc])
    return theta_list


class DTMTrainer:
    def __init__(self,
                 dataset,
                 num_topics=50,
                 num_top_words=15,
                 alphas=0.01,
                 chain_variance=0.005,
                 passes=10,
                 lda_inference_max_iter=25,
                 em_min_iter=6,
                 em_max_iter=20,
                 verbose=False
                ):

        self.dataset = dataset
        self.vocab_size = dataset.vocab_size
        self.num_topics = num_topics
        self.num_top_words = num_top_words
        self.alphas = alphas
        self.chain_variance = chain_variance
        self.passes = passes
        self.lda_inference_max_iter = lda_inference_max_iter
        self.em_min_iter = em_min_iter
        self.em_max_iter = em_max_iter

        self.verbose = verbose
        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

    def train(self):
        id2word = dict(zip(range(self.vocab_size), self.dataset.vocab))
        train_bow = self.dataset.train_bow
        train_times = self.dataset.train_times.astype('int32')

        # order documents by time slices
        self.doc_order_idx = np.argsort(train_times)
        train_bow = train_bow[self.doc_order_idx]
        time_slices = np.bincount(train_times)

        corpus = gensim.matutils.Dense2Corpus(train_bow, documents_columns=False)

        self.model = ldaseqmodel.LdaSeqModel(
            corpus=corpus,
            id2word=id2word,
            time_slice=time_slices,
            num_topics=self.num_topics,
            alphas=self.alphas,
            chain_variance=self.chain_variance,
            em_min_iter=self.em_min_iter,
            em_max_iter=self.em_max_iter,
            lda_inference_max_iter=self.lda_inference_max_iter,
            passes=self.passes
        )

    def test(self, bow):
        # bow = dataset.bow.cpu().numpy()
        # times = dataset.times.cpu().numpy()
        corpus = gensim.matutils.Dense2Corpus(bow, documents_columns=False)

        num_workers = 20
        split_idx_list = np.array_split(np.arange(len(bow)), num_workers)
        worker_size_list = [len(x) for x in split_idx_list]

        worker_id = 0
        docs_list = [list() for i in range(num_workers)]
        for i, doc in enumerate(corpus):
            docs_list[worker_id].append(doc)
            if len(docs_list[worker_id]) >= worker_size_list[worker_id]:
                worker_id += 1

        args_list = list()
        for docs in docs_list:
            args_list.append([self.model, docs])

        starttime = datetime.datetime.now()

        pool = Pool(processes=num_workers)
        results = pool.map(work, args_list)

        pool.close()
        pool.join()

        theta_list = list()
        for rst in results:
            theta_list.extend(rst)

        endtime = datetime.datetime.now()

        print("DTM test time: {}s".format((endtime - starttime).seconds))

        return np.asarray(theta_list)

    def get_theta(self):
        theta = self.model.gammas / self.model.gammas.sum(axis=1)[:, np.newaxis]
        # NOTE: MUST transform gamma to original order.
        return theta[np.argsort(self.doc_order_idx)]

    def get_beta(self):
        beta = list()
        # K x V x T
        for item in self.model.topic_chains:
            # V x T
            beta.append(item.e_log_prob)

        # T x K x V
        beta = np.transpose(np.asarray(beta), (2, 0, 1))
        # use softmax
        beta = np.exp(beta)
        beta = beta / beta.sum(-1, keepdims=True)
        return beta

    def get_top_words(self, num_top_words=None):
        if num_top_words is None:
            num_top_words = self.num_top_words
        beta = self.get_beta()
        top_words_list = list()
        for time in range(beta.shape[0]):
            top_words = _utils.get_top_words(beta[time], self.dataset.vocab, num_top_words, self.verbose)
            top_words_list.append(top_words)
        return top_words_list

    def export_theta(self):
        train_theta = self.get_theta()
        test_theta = self.test(self.dataset.test_bow)
        return train_theta, test_theta
